# coding=utf-8
"""
"""
from __future__ import absolute_import, print_function, division

from datetime import datetime, timedelta
import sqlalchemy as sa

from flask import render_template, current_app

from abilian.i18n import _l
from abilian.core.models.subjects import User
from abilian.services.audit import AuditEntry, CREATION

from ..panel import AdminPanel


class DashboardPanel(AdminPanel):
  id = 'dashboard'
  path = ''
  label = _l('Dashboard')
  icon = 'eye-open'

  def get(self):
    stats = {
      'today': stats_since(timedelta(days=1)),
      'this_week': stats_since(timedelta(days=7)),
      'this_month': stats_since(timedelta(days=30)),
    }
    return render_template("admin/dashboard.html", stats=stats)


def stats_since(dt):
  new_members = new_documents = new_messages = 0
  after_date = datetime.utcnow() - dt
  session = current_app.db.session()
  counts_per_type = session\
    .query(AuditEntry.entity_type.label('type'),
           sa.func.count(AuditEntry.entity_type).label('count'))\
    .group_by(AuditEntry.entity_type)\
    .filter(AuditEntry.happened_at > after_date) \
    .filter(AuditEntry.type == CREATION)\
    .all()

  for entity_type, count in counts_per_type:
    entity_class = entity_type.split('.')[-1]
    if entity_class == 'User':
      new_members = count
    elif entity_class == 'Document':
      new_documents = count
    elif entity_class == 'Message':
      new_messages = count

  active_users = session.query(sa.func.count(User.id))\
                        .filter(User.last_active > after_date).scalar()

  return {
    'new_members': new_members,
    'active_users': active_users,
    'new_documents': new_documents,
    'new_messages': new_messages,
  }



# coding=utf-8
"""
Various tools that don't belong some place specific.
"""
from __future__ import absolute_import, print_function, division

import sys
import functools
import logging
import time
import pytz
from math import ceil
import unicodedata
import re

from flask import request, current_app
from babel.dates import LOCALTZ
from werkzeug.local import LocalProxy

try:
  import ipdb as pdb
except ImportError:
  import pdb

if sys.version_info[0] > 2:
  unicode = str


def pdb_on_error(fn):
  """
  Decorator to trigger (i)pdb on exception inside decorated function.
  Active only in DEBUG mode.
  Useful to debug POST only views for example.
  """
  @functools.wraps(fn)
  def decorator(*args, **kwargs):
    try:
      return fn(*args, **kwargs)
    except:
      if current_app and current_app.debug:
        type_, value, tb = sys.exc_info()
        if tb.tb_next is not None:
          # error has happened inside decorated function, remove us from top
          # stack: better readability in logs, accurate label in sentry
          tb = tb.tb_next

        pdb.post_mortem(tb)
      raise

  return decorator


def noproxy(obj):
  """
  Unwrap obj from werkzeug.local.LocalProxy if needed. This is required if
  one want to test `isinstance(obj, SomeClass)`.
  """
  if isinstance(obj, LocalProxy):
    obj = obj._get_current_object()
  return obj


def fqcn(cls):
  """
  Fully Qualified Class Name
  """
  return unicode(cls.__module__ + '.' + cls.__name__)


def friendly_fqcn(cls_name):
  """
  Friendly name of fully qualified class name.
  :param cls_name: a string or a class
  """
  if isinstance(cls_name, type):
    cls_name = fqcn(cls_name)

  return cls_name.rsplit('.', 1)[-1]


def local_dt(dt):
  """ Return an aware datetime in system timezone, from a naive or aware
  datetime. Naive datetime are assumed to be in UTC TZ.
  """
  if not dt.tzinfo:
    dt = pytz.utc.localize(dt)
  return LOCALTZ.normalize(dt.astimezone(LOCALTZ))


def utc_dt(dt):
  """ set UTC timezone on a datetime object. A naive datetime is assumed to be
  in UTC TZ
  """
  if not dt.tzinfo:
    return pytz.utc.localize(dt)
  return dt.astimezone(pytz.utc)


def get_params(names):
  """
  Returns a dictionary with params from request.

  TODO: I think we don't use it anymore and it should be removed before
  someone gets hurt.
  """
  params = {}
  for name in names:
    value = request.form.get(name) or request.files.get(name)
    if value is not None:
      params[name] = value
  return params


class timer(object):
  """
  Decorator that mesures the time it takes to run a function.
  """
  __instances = {}

  def __init__(self, f):
    self.__f = f
    self.log = logging.getLogger(f.__module__ + '.' + f.__name__)

  def __call__(self, *args, **kwargs):
    self.__start = time.time()
    result = self.__f(*args, **kwargs)
    value = time.time() - self.__start
    self.log.info('elapsed time: {0:.2f}ms'.format(value * 1000))
    return result


# From http://wiki.python.org/moin/PythonDecoratorLibrary#Memoize
class memoized(object):
  """
  Decorator. Caches a function's return value each time it is called.
  If called later with the same arguments, the cached value is returned
  (not reevaluated).
  """

  def __init__(self, func):
    self.func = func
    self.cache = {}
    functools.wraps(func)(self)

  def __call__(self, *args):
    try:
      return self.cache[args]
    except KeyError:
      value = self.func(*args)
      self.cache[args] = value
      return value
    except TypeError:
      # uncachable -- for instance, passing a list as an argument.
      # Better to not cache than to blow up entirely.
      return self.func(*args)

  def __get__(self, obj, objtype):
    """Support instance methods."""
    return functools.partial(self.__call__, obj)


# From http://flask.pocoo.org/snippets/44/
class Pagination(object):

  def __init__(self, page, per_page, total_count):
    self.page = page
    self.per_page = per_page
    self.total_count = total_count

  @property
  def pages(self):
    return int(ceil(self.total_count / float(self.per_page)))

  @property
  def has_prev(self):
    return self.page > 1

  @property
  def prev(self):
    return self.page - 1 if self.has_prev else None

  @property
  def has_next(self):
    return self.page < self.pages

  @property
  def next(self):
    return self.page + 1 if self.has_next else None

  def iter_pages(self, left_edge=2, left_current=2,
                 right_current=5, right_edge=2):
    last = 0
    for num in range(1, self.pages + 1):
      if (num <= left_edge or
          (self.page - left_current - 1 < num < self.page + right_current) or
          num > self.pages - right_edge):
        if last + 1 != num:
          yield None
        yield num
        last = num

_NOT_WORD_RE = re.compile(r'[^\w\s]+', flags=re.UNICODE)


def slugify(value, separator=u"-"):
  """
    Slugify an unicode string, to make it URL friendly.
  """
  value = unicode(value)
  separator = unicode(separator)
  value = _NOT_WORD_RE.sub(u' ', value)
  value = unicodedata.normalize('NFKD', value)
  value = value.encode('ascii', 'ignore')
  value = value.decode('ascii')
  value = value.strip().lower()
  value = re.sub(r'[%s_\s]+' % separator, separator, value)
  return value


class BasePresenter(object):
  """
  A presenter wraps a model an adds specific (often, web-centric) accessors.
  subclass to make it useful. Presenters are immutable.
  """
  def __init__(self, model):
    self._model = model

  def __getattr__(self, key):
    return getattr(self._model, key)

  def __setattr__(self, key, value):
    """Make presenter immutable.
    """
    if key == '_model':
      self.__dict__[key] = value
    else:
      raise AttributeError("Can't set attribute on a presenter.")

  @classmethod
  def wrap_collection(cls, models):
    return [cls(model) for model in models]



"""
Add a few specific filters to Jinja2.
"""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import re
from functools import wraps
import datetime
from calendar import timegm

from flask import Flask
from flask.ext import babel
from jinja2 import Markup, escape, evalcontextfilter
from pytz import utc
from babel.dates import DateTimePattern, format_timedelta, parse_pattern
import bleach
from werkzeug.routing import BuildError

from ..core.util import local_dt, utc_dt, slugify
from .util import url_for


def autoescape(filter_func):
  """
  Decorator to autoescape result from filters.
  """
  @evalcontextfilter
  @wraps(filter_func)
  def _autoescape(eval_ctx, *args, **kwargs):
    result = filter_func(*args, **kwargs)
    if eval_ctx.autoescape:
      result = Markup(result)
    return result
  return _autoescape


@autoescape
def nl2br(value):
  """
  Replace newlines with <br />.
  """
  result = escape(value).replace(u'\n', Markup(u'<br />\n'))
  return result

_PARAGRAPH_RE = re.compile(r'(?:\r\n|\r|\n){2,}')


@autoescape
def paragraphs(value):
  """
  Blank lines delimitates paragraphs.
  """
  result = u'\n\n'.join(
    (u'<p>{}</p>'.format(p.strip().replace('\n', Markup('<br />\n')))
     for p in _PARAGRAPH_RE.split(escape(value))))
  return result


def labelize(s):
  return u" ".join([ w.capitalize() for w in s.split(u"_") ])


def filesize(d):
  if not isinstance(d, int):
    d = int(d)

  if d < 1000:
    s = "%d&nbsp;B" % d

  elif d < 1e4:
    s = "%.1f&nbsp;kB" % (d / 1e3)
  elif d < 1e6:
    s = "%.0f&nbsp;kB" % (d / 1e3)

  elif d < 1e7:
    s = "%.1f&nbsp;MB" % (d / 1e6)
  elif d < 1e9:
    s = "%.0f&nbsp;MB" % (d / 1e6)

  elif d < 1e10:
    s = "%.1f&nbsp;GB" % (d / 1e9)

  else:
    s = "%.0f&nbsp;GB" % (d / 1e9)

  return Markup(s)

def roughsize(size, above=20, mod=10):
  """
  6 -> '6'
  15 -> '15'
  134 -> '130+'
  """
  if size < above:
    return unicode(size)

  return u'{:d}+'.format(size - size % mod)


def age(dt, now=None):
  # Fail silently for now XXX
  if not dt:
    return ""

  if not now:
    now = datetime.datetime.utcnow()

  dt = utc_dt(dt)
  now = utc_dt(now)

  # don't use (flask.ext.)babel.format_timedelta: as of Flask-Babel 0.9 it
  # doesn't support "threshold" arg.
  return format_timedelta((dt - now),
                          locale=babel.get_locale(),
                          granularity='minute',
                          threshold=0.9,
                          add_direction=True)


def date_age(dt, now=None):
  # Fail silently for now XXX
  if not dt:
    return ""

  formatted_date = babel.format_datetime(dt, format='yyyy-MM-dd HH:mm')
  return u"{} ({})".format(formatted_date, age(dt, now))


def date(value, format="EE, d MMMM y"):
  if isinstance(value, datetime.date):
    return babel.format_date(value, format)
  else:
    return babel.format_date(local_dt(value), format)


def babel2datepicker(pattern):
  """
  Converts date format from babel
  (http://babel.pocoo.org/docs/dates/#date-fields)) to a format understood by
  bootstrap-datepicker.
  """
  if not isinstance(pattern, DateTimePattern):
    pattern = parse_pattern(pattern)

  map_fmt = {
    # days
    'd': 'dd',
    'dd': 'dd',
    'EEE': 'D',
    'EEEE': 'DD',
    'EEEEE': 'D',  # narrow name => short name
    # months
    'M': 'mm',
    'MM': 'mm',
    'MMM': 'M',
    'MMMM': 'MM',
    # years
    'y': 'yyyy',
    'yy': 'yyyy',
    'yyy': 'yyyy',
    'yyyy': 'yyyy',

    # time picker format
    # hours
    'h': '%I',
    'hh': '%I',
    'H': '%H',
    'HH': '%H',
    # minutes,
    'm': '%M',
    'mm': '%M',
    # seconds
    's': '%S',
    'ss': '%S',
    # am/pm
    'a': '%p',
  }

  return pattern.format % map_fmt


# Doesn't work yet. TZ issues.
def to_timestamp(dt):
  utc_datetime = dt.astimezone(utc)
  return timegm(utc_datetime.timetuple()) + utc_datetime.microsecond / 1e6


def abbrev(s, max_size):
  if len(s) <= max_size:
    return s
  else:
    h = max_size // 2 - 1
    return s[0:h] + "..." + s[-h:]


@autoescape
def linkify(s):
  return Markup(bleach.linkify(s))


def obj_to_url(obj):
  """
  Find url for obj using :func:`url_for`, return empty string is not found.

  :func:`url_for` is also provided in jinja context, the filtering version is
  forgiving when `obj` has no default view set.
  """
  try:
    return url_for(obj)
  except BuildError:
    return u''


def init_filters(env):
  if isinstance(env, Flask):
    # old api for init_filters: we used to pass flask application
    env = env.jinja_env

  env.filters['nl2br'] = nl2br
  env.filters['paragraphs'] = paragraphs
  env.filters['date_age'] = date_age
  env.filters['age'] = age
  env.filters['date'] = date
  env.filters['babel2datepicker'] = babel2datepicker
  env.filters['to_timestamp'] = to_timestamp
  env.filters['url_for'] = obj_to_url
  env.filters['abbrev'] = abbrev
  env.filters['filesize'] = filesize
  env.filters['roughsize'] = roughsize
  env.filters['labelize'] = labelize
  env.filters['linkify'] = linkify
  env.filters['toslug'] = slugify



# coding=utf-8
"""
"""
from __future__ import absolute_import, print_function, division
from datetime import datetime

from flask import request
from flask_login import current_user
from flask_sqlalchemy import BaseQuery
from sqlalchemy import ForeignKey, Column, DateTime, Integer, String
from sqlalchemy.orm import relation

from abilian.core.extensions import db
from abilian.core.models.subjects import User


_MARK = object()


class LoginSessionQuery(BaseQuery):

  def get_active_for(self, user, user_agent=_MARK, ip_address=_MARK):
    """
    Returns last known session for user

    :param user: user session
    :type user: `abilian.core.models.subjects.User`

    :param user_agent: *exact* user agent string to lookup, or `None` to have
    user_agent extracted from request object. If not provided at all, no
    filtering on user_agent.
    :type user_agent: string or None, or absent

    :param ip_address: client IP, or `None` to have ip_address extracted from
    request object (requires header 'X-Forwarded-For'). If not provided at
    all, no filtering on ip_address.
    :type ip_address: string or None, or absent

    :rtype: `LoginSession` or `None` if no session is found.
    """
    conditions = [LoginSession.user == user]

    if user_agent is not _MARK:
      if user_agent is None:
        user_agent = request.environ.get('HTTP_USER_AGENT', '')
      conditions.append(LoginSession.user_agent == user_agent)

    if ip_address is not _MARK:
      if ip_address is None:
        ip_address = request.headers.getlist("X-Forwarded-For")
        ip_address = ip_address[0] if ip_address else request.remote_addr
      conditions.append(LoginSession.ip_address == ip_address)

    session = LoginSession.query\
      .filter(*conditions)\
      .order_by(LoginSession.id.desc())\
      .first()
    return session


class LoginSession(db.Model):

  __tablename__ = 'login_session'
  query_class = LoginSessionQuery

  id = Column(Integer, primary_key=True, autoincrement=True)

  started_at = Column(DateTime, default=datetime.utcnow, nullable=False)
  last_active_at = Column(DateTime, default=datetime.utcnow, nullable=False)
  ended_at = Column(DateTime)

  user_id = Column(ForeignKey(User.id), nullable=False)
  user = relation(User)

  #: IP addres, should work with both IPv4 and IPv6.
  # FIXME: define a custom type to use postgres inet when on postgres
  # http://www.postgresql.org/docs/9.1/static/datatype-net-types.html
  ip_address = Column(String(39), default="", nullable=False)

  #: Limited (somewhat artbitrarily) to 512 characters.
  user_agent = Column(String(512), default="", nullable=False)

  @staticmethod
  def new():
    user_agent = request.environ.get('HTTP_USER_AGENT', '')
    if not request.headers.getlist("X-Forwarded-For"):
      ip_address = request.remote_addr
    else:
      ip_address = request.headers.getlist("X-Forwarded-For")[0]
    session = LoginSession(user=current_user, user_agent=user_agent,
                           ip_address=ip_address)
    return session



# coding=utf-8
"""
"""
from __future__ import absolute_import, print_function, division

from werkzeug.exceptions import BadRequest
from flask import current_app

from abilian.i18n import _l
from abilian.core.entities import Entity
from abilian.core.models.tag import Tag
from abilian.web import url_for
from abilian.web.blueprints import Blueprint
from abilian.web.views import (
  JSONView, BaseObjectView, ObjectEdit, ObjectCreate, ObjectDelete,
)

from .forms import TagForm

bp = Blueprint('tags', __name__, url_prefix='/tags')


class BaseTagView(object):
  """
  Mixin for tag views
  """
  Model = Tag
  Form = TagForm

  def __init__(self, *args, **kwargs):
    super(BaseTagView, self).__init__(*args, **kwargs)
    self.extension = current_app.extensions['tags']

  def view_url(self):
    return url_for()


class TagEdit(BaseTagView, ObjectEdit):
  _message_success = _l(u'Tag edited')


edit_view = TagEdit.as_view('edit')
bp.route('/manage/<int:object_id>/edit')(edit_view)


class TagCreate(BaseTagView, ObjectCreate):
  _message_success = _l(u'Tag created')


create_view = TagCreate.as_view('create')
bp.route('/manage/new')(create_view)


class TagDelete(BaseTagView, ObjectDelete):
  _message_success = _l(u'Tag deleted')


delete_view = TagDelete.as_view('delete')
bp.route('/manage/<int:object_id>/delete')(delete_view)


# Tags on entities
entity_bp = Blueprint('entity_tags', __name__, url_prefix='/tags/entity')


class BaseEntityTagView(BaseTagView):

  def init_object(self, args, kwargs):
    args, kwargs = super(BaseEntityTagView, self).init_object(args, kwargs)
    entity_id = kwargs.pop('entity_id', None)

    if entity_id is not None:
      self.entity = Entity.query.get(entity_id)

    if self.entity is None:
      raise BadRequest('No entity provided')

    return args, kwargs

  def view_url(self):
    return url_for(self.entity)

  def index_url(self):
    return self.view_url()


class EntityTagList(BaseEntityTagView, BaseObjectView, JSONView):

  def get(self, *args, **kwargs):
    return JSONView.get(self, *args, **kwargs)

  def data(self, *args, **kwargs):
    tags = sorted(self.extension.entity_tags(self.entity))
    return dict(result=tags)


entity_bp.route('/<int:object_id>/list')(EntityTagList.as_view('list'))


class EntityTagManage(BaseEntityTagView, ObjectEdit):
  methods = ['POST']

  # operation: add or remove
  mode = None

  def __init__(self, mode, *args, **kwargs):
    super(EntityTagManage, *args, **kwargs)
    assert mode in ('add', 'remove')
    self.mode = mode

  def form_valid(self):
    ns = self.form.ns.data
    label = self.form.label.data
    op = getattr(self.extension, self.mode)
    op(self.entity, ns=ns, label=label)


entity_bp.route('/<int:object_id>/add')(EntityTagManage.as_view('add', mode='add'))
entity_bp.route('/<int:object_id>/remove')(EntityTagManage.as_view('remove', mode='remove'))


class EntityTagEdit(ObjectEdit):
  Model = Entity

  def init_object(self, args, kwargs):
    args, kwargs = super(EntityTagEdit, self).init_object(args, kwargs)
    extension = current_app.extensions['tags']
    self.Form = extension.entity_tags_form(self.obj)
    return args, kwargs

  def view_url(self):
    return url_for(self.obj)

  def index_url(self):
    return self.view_url()


entity_bp.route('/<int:object_id>/edit')(EntityTagEdit.as_view('edit'))
